#pragma once

#include <memory>

#include <amdis/Output.hpp>

namespace AMDiS
{
  /**
   * \addtogroup Adaption
   * @{
   **/

  typedef enum {
    NO_OPERATION = 0,
    INTERPOLATE  = 1
  } DataTransferOperation;


  /// \brief Interface for Containers allowing data transfer between grid changes.
  template <class Container>
  class DataTransferInterface
  {
  public:
    /// Virtual destructor
    virtual ~DataTransferInterface() = default;

    /// Collect data that is needed before grid adaption
    virtual void preAdapt(Container const& container, bool mightCoarsen) = 0;

    /// Interpolate data to new grid after grid adaption
    virtual void postAdapt(Container& container, bool refined) = 0;
  };


  /**
   * \brief Implementation of \ref DataTransferInterface that does not interpolate, but
   * just resizes the containers to the dimension of the basis
   **/
  template <class Container>
  class NoDataTransfer
      : public DataTransferInterface<Container>
  {
  public:
    void preAdapt(Container const& container, bool) override {}

    void postAdapt(Container& container, bool) override
    {
      container.compress();
    }
  };


  template <class Container, class Basis>
  class DataTransfer;

  /// Factory to create DataTransfer objects based on the \ref DataTransferOperation
  template <class Container>
  class DataTransferFactory
  {
    using Interface = DataTransferInterface<Container>;

  public:
    template <class Basis>
    static std::unique_ptr<Interface> create(DataTransferOperation op, Basis const& basis)
    {
      switch (op)
      {
        case NO_OPERATION:
          return std::make_unique<NoDataTransfer<Container>>();
        case INTERPOLATE:
          return std::make_unique< DataTransfer<Container, Basis> >(basis);
        default:
          error_exit("Invalid data transfer\n");
          return nullptr; // avoid warnings
      }
    }
  };

  /// @}

} // end namespace AMDiS

#include "DataTransfer.inc.hpp"
