#pragma once

#include <amdis/Assembler.hpp>
#include <amdis/LocalOperator.hpp>
#include <amdis/typetree/Traversal.hpp>
#include <amdis/utility/AssembleOperators.hpp>

namespace AMDiS {

template <class GB, class B>
void DOFVectorBase<GB,B>::
init(bool asmVector)
{
  backend_.resize(size());
  if (asmVector)
    backend_.set(0);

  auto const localSize = basis_->localView().maxSize();
  elementVector_.resize(localSize);
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
finish(bool /*asmVector*/)
{}


template <class GB, class B>
void DOFVectorBase<GB,B>::
insert(LocalView const& localView, ElementVector const& elementVector)
{
  for (size_type i = 0; i < localView.size(); ++i) {
    if (std::abs(elementVector[i]) > threshold<double>) {
      size_type const idx = flatMultiIndex(localView.index(i));
      backend_[idx] += elementVector[i];
    }
  }
}


template <class GB, class B>
  template <class ContextTag, class Expr, class TreePath>
void DOFVectorBase<GB,B>::
addOperator(ContextTag contextTag, Expr const& expr, TreePath path)
{
  static_assert( Concepts::PreTreePath<TreePath>,
      "path must be a valid treepath, or an integer/index-constant");

  auto i = child(basis_->localView().tree(), makeTreePath(path));

  using LocalContext = typename ContextTag::type;
  auto op = makeLocalOperator<LocalContext>(expr, basis_->gridView());
  auto localAssembler = makeUniquePtr(makeAssembler<LocalContext>(std::move(op), i));

  operators_[i].push(contextTag, std::move(localAssembler));
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
assemble(LocalView const& localView)
{
  elementVector_ = 0;
  auto const& gv = basis_->gridView();
  auto const& element = localView.element();
  auto geometry = element.geometry();

  for_each_node(localView.tree(), [&](auto const& node, auto) {
    auto& rhsOp = operators_[node];
    if (rhsOp) {
      rhsOp.bind(element, geometry);
      assembleOperators(gv, element, rhsOp, makeVectorAssembler(node, elementVector_));
      rhsOp.unbind();
    }
  });

  insert(localView, elementVector_);
}


template <class GB, class B>
void DOFVectorBase<GB,B>::
assemble()
{
  auto localView = basis_->localView();

  init(true);
  for (auto const& element : elements(basis_->gridView())) {
    localView.bind(element);
    assemble(localView);
    localView.unbind(element);
  }
  finish(true);
}

} // end namespace AMDiS
